#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""keywords.py

Keywords class and keywords() method

"""
# logging
import logging

# requirements
from pandas import DataFrame

# part of module
from .cache import generate_idx
from .counts import score_counts

logger = logging.getLogger(__name__)


class Keywords:
    """ keyword analysis """

    def __init__(self, corpus, df_dump, p_query=['lemma'],
                 counts=None):

        # consistency check
        if df_dump is not None and len(df_dump) == 0:
            logger.warning('empty dump')
            return

        # init corpus
        self.corpus = corpus

        # determine layer to work on
        self.p_query = [p_query] if isinstance(p_query, str) else p_query
        available_attributes = self.corpus.available_attributes()['attribute'].values
        if not set(self.p_query).issubset(set(available_attributes)):
            logger.warning('specfied p-attribute(s) ("%s") not available' % " ".join(self.p_query))
            logger.warning('falling back to primary layer')
            self.p_query = ['word']

        # collect context and save result
        if df_dump is not None:

            # get from cache if possible
            identifier = generate_idx([df_dump.reset_index()[['match', 'matchend']], p_query])
            counts = self.corpus.cache.get(identifier + "-matchcounts")

            # create and cache otherwise
            if not isinstance(counts, DataFrame):
                logger.info('collecting token counts of subcorpus')
                counts = corpus.counts.dump(
                    df_dump, start='match', end='matchend', p_atts=self.p_query, split=True
                )
                self.corpus.cache.set(identifier + "-matchcounts", counts)

        else:
            if counts is None:
                logger.error('if no dump is given, you have to provide all counts')
                return

        self.counts = counts

    def show(self, order='log_likelihood', cut_off=100, ams=None,
             min_freq=2, flags=None, marginals='corpus',
             show_negative=False):

        # consistency check
        if self.counts.empty:
            logger.warning("nothing to show")
            return DataFrame()

        # get subcorpus frequencies und apply min freq
        f = self.counts.rename(columns={'freq': 'f'})
        vocab = len(f)
        f1 = f['f'].sum()
        f = f.loc[f['f'] >= min_freq]

        # get reference frequency
        if isinstance(marginals, str):
            if marginals == 'corpus':
                N = self.corpus.corpus_size
                marginals = self.corpus.marginals(f.index, self.p_query)
            else:
                raise NotImplementedError
        elif isinstance(marginals, DataFrame):
            # DataFrame must contain a column 'freq'
            N = marginals['freq'].sum()
        else:
            raise NotImplementedError

        # create dataframe
        f2 = marginals[['freq']].rename(columns={'freq': 'f2'})
        df = f.join(f2, how='left')
        df = df.fillna(0)
        df['f1'] = f1
        df['N'] = N

        # score
        keywords = score_counts(df, order=order, cut_off=cut_off,
                                flags=flags, ams=ams, vocab=vocab)

        # throw away anti-keywords by default
        if not show_negative:
            keywords = keywords.loc[keywords['O11'] >= keywords['E11']]

        return keywords


def keywords(corpus, corpus_reference, p=['lemma'],
             p_reference=['lemma'], order='O11', cut_off=100,
             ams=None, min_freq=2, flags=None):
    """directly extract keywords by comparing two (sub-)corpora

    :param ccc.Corpus corpus: target corpus
    :param ccc.Corpus corpus_reference: reference corpus
    :param list p: p-att (str possible)
    :param list p_reference: p-att (str possible)
    :param str order: AM to order
    :param list ams: list of AMs
    :param int min_freq: minimum number of occurrences in target corpus
    :param list flags: '%cd'

    """

    # get from cache if possible
    identifier = generate_idx([corpus.__str__(), corpus_reference.__str__(),
                               p, p_reference, order, cut_off, ams, min_freq, flags], prefix='kw-')
    keywords = corpus.cache.get(identifier)

    if keywords is None:
        # get and merge both dataframes of counts
        logger.info('starting keyword analysis')
        target = corpus.marginals(p_atts=p)[['freq']].rename(columns={'freq': 'f1'})
        reference = corpus_reference.marginals(p_atts=p_reference)[['freq']].rename(columns={'freq': 'f2'})

        logger.info('-- combining frequency lists')
        df = target.join(reference, how='outer')
        df['N1'] = target['f1'].sum()
        df['N2'] = reference['f2'].sum()
        df = df.fillna(0)  # , downcast='infer')

        logger.info('-- cut-off')
        vocab = len(df)
        df = df.loc[df['f1'] >= min_freq]

        # score counts
        keywords = score_counts(df, order=order, cut_off=cut_off,
                                flags=flags, ams=ams, digits=4, vocab=vocab)
        logger.info('-- done with keyword analysis')
        corpus.cache.set(identifier, keywords)

    return keywords



import pytest
from pandas import DataFrame

from ccc.cwb import Corpus
from ccc.dumps import Dumps

from .conftest import DATA_PATH


def get_corpus(corpus_settings, data_dir=DATA_PATH):

    return Corpus(
        corpus_settings['corpus_name'],
        registry_dir=corpus_settings['registry_dir'],
        lib_dir=corpus_settings.get('lib_dir', None),
        data_dir=data_dir
    )


########
# DUMP #
########
def test_query(germaparl):
    corpus = get_corpus(germaparl)
    subcorpus = corpus.query('"SPD"')
    assert subcorpus.subcorpus_name == 'df_b37e127b31'
    assert len(subcorpus.df) == 632
    assert (subcorpus.df.columns == ['context', 'contextend']).all()


def test_query_name(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"', name="SPD")
    assert dump.subcorpus_name == "SPD"
    assert len(dump.df) == 632
    assert (dump.df.columns == ['context', 'contextend']).all()


def test_breakdown(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')
    breakdown = dump.breakdown()
    assert isinstance(breakdown, DataFrame)
    assert breakdown.loc['SPD']['freq'] == 632


def test_breakdown_p_att(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('[lemma="gehen"]')
    breakdown = dump.breakdown()
    assert isinstance(breakdown, DataFrame)
    assert breakdown.loc['geht']['freq'] == 150
    breakdown = dump.breakdown(p_atts=['lemma'])
    assert isinstance(breakdown, DataFrame)
    assert breakdown.loc['gehen']['freq'] == 224


def test_matches(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')
    matches = dump.matches()
    assert isinstance(matches, set)
    assert 8193 in matches


def test_matches_subcorpus(germaparl):
    corpus = get_corpus(germaparl)
    dump_base = corpus.query(r'[pos="NE"]? [pos="NE"] "\[" ".*" "\]"', name="Base")
    tokens_base = len(dump_base.matches())
    dump_neg = dump_base.query('[pos="NE"]')
    tokens_neg = len(dump_neg.matches())
    assert (tokens_base - tokens_neg) == 350


def test_context(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')
    context = dump.context()
    assert isinstance(context, DataFrame)
    assert context['context'][0] == 76


def test_set_context(germaparl):

    corpus = get_corpus(germaparl)
    dump = corpus.query('"CSU"', context_break='text')
    assert dump.df['context'].iloc[0] == 620
    dump = dump.set_context(10, context_break='text', context_right=10)
    assert dump.df['context'].iloc[0] == 630


def test_query_s_satt(germaparl):
    corpus = get_corpus(germaparl)
    parties = {"GRUENE", "Bündnis 90/Die Grünen"}
    dump = corpus.query_s_att('text_party', parties)
    assert dump.subcorpus_name is not None
    assert len(dump.df) == 87
    assert (dump.df.columns == ['text_party_cwbid', 'text_party']).all()


def test_concordance(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')
    conc = dump.concordance()
    assert isinstance(conc, DataFrame)
    assert conc['word'].iloc[0].startswith("früh")


def test_concordance_options(germaparl):

    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')

    conc = dump.concordance(form='simple')
    assert isinstance(conc, DataFrame)
    assert len(conc) == 100
    assert 'word' in conc.columns
    conc = dump.concordance(form='kwic')
    assert isinstance(conc, DataFrame)
    assert len(conc) == 100
    assert 'node_word' in conc.columns
    conc = dump.concordance(form='dict')
    assert isinstance(conc, DataFrame)
    assert len(conc) == 100
    assert 'dict' in conc.columns
    conc = dump.concordance(form='slots')
    assert isinstance(conc, DataFrame)
    assert len(conc) == 100
    assert 'match..matchend_word' in conc.columns
    conc = dump.concordance(form='dataframe')
    assert isinstance(conc, DataFrame)
    assert len(conc) == 100
    assert 'dataframe' in conc.columns


def test_collocates(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')
    coll = dump.collocates()
    assert coll.index[0] == 'die'


def test_collocates_options(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')
    coll = dump.collocates(order='log_likelihood', cut_off=70)
    assert coll.index[0] == 'bei'
    assert len(coll) == 70


def test_keywords(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD" expand to s')
    kw = dump.keywords()
    assert kw.index[0] == 'die'


def test_keywords_options(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD" expand to s')
    kw = dump.keywords(order='log_ratio', cut_off=200)
    assert kw.index[0] == 'SPD'


#########
# DUMPS #
#########
@pytest.mark.dumps
def test_dumps_keywords(germaparl):

    # subcorpora via s-attribute values
    parties = {
        'green': {"GRUENE", "Bündnis 90/Die Grünen"},
        'red': {'SPD'},
        'black': {'CDU', 'CSU'},
        'yellow': {'FDP'},
        'purple': {'PDS'}
    }

    # keywords
    corpus = get_corpus(germaparl)
    dumps = Dumps(corpus, parties, s_att='text_party')
    table = dumps.keywords(order='log_ratio')
    assert table.loc['green'].index[0] == "Oppositionsfraktion"
    assert table.loc['red'].index[0] == "Bereicherung"
    assert table.loc['black'].index[0] == "Abgabenquote"
    assert table.loc['yellow'].index[0] == "Wirtschafts-"
    assert table.loc['purple'].index[0] == "ÖPNV-Gesetz"


@pytest.mark.dumps
def test_dumps_collocates(germaparl):

    # subcorpora via s-attribute values
    parties = {
        'green': {"GRUENE", "Bündnis 90/Die Grünen"},
        'red': {'SPD'},
        'black': {'CDU', 'CSU'},
        'yellow': {'FDP'},
        'purple': {'PDS'}
    }

    # collocates
    corpus = get_corpus(germaparl)
    dumps = Dumps(corpus, parties, s_att='text_party')
    table = dumps.collocates(
        cqp_query='"Wirtschaft"',
        order='log_ratio',
        context_break='s',
        window=20
    )
    assert len(set(table.index.get_level_values('subcorpus'))) == len(parties)
    assert table.loc['yellow'].index[0] == 'Grad'


@pytest.mark.dumps
def test_dumps_collocates_global(germaparl):

    # subcorpora via s-attribute values
    parties = {
        'green': {"GRUENE", "Bündnis 90/Die Grünen"},
        'red': {'SPD'},
        'black': {'CDU', 'CSU'},
        'yellow': {'FDP'},
        'purple': {'PDS'}
    }

    # collocates
    corpus = get_corpus(germaparl)
    dumps = Dumps(corpus, parties, s_att='text_party')
    table = dumps.collocates(
        cqp_query='"Wirtschaft"',
        order='log_ratio',
        context_break='s',
        window=20,
        marginals='corpus'
    )
    assert len(set(table.index.get_level_values('subcorpus'))) == len(parties)
    assert table.loc['yellow'].index[0] == 'Grad'


@pytest.mark.benchmark
def test_perf_dumps(benchmark, germaparl):
    benchmark.pedantic(test_dumps_collocates, kwargs={'germaparl': germaparl}, rounds=5, iterations=2)



import pytest
from pandas import DataFrame

from ccc import Corpus
from ccc.discoursemes import (Constellation, TextConstellation,
                              create_constellation)
from ccc.utils import format_cqp_query

from .conftest import DATA_PATH

#######################
# ccc.discoursemes ####
#######################

# constellation_merge
# role_formatter
# calculate_collocates

# Constellation
# .__init__
# .add_discourseme: constellation_merge
# .group_lines
# .concordance: role_formatter, .group_lines
# .collocates: calculate_collocates

# create_constellation

############
# mmda.ccc #
############

# - ccc_concordance:
#   create_constellation, Constellation.concordance, role_formatter
# - ccc_collocates:
#   create_constellation, Constellation.collocates

# - ccc_constellation_concordance
#   get_constellation_df
# - ccc_constellation_association
#   get_constellation_df


def get_corpus(corpus_settings, data_dir=DATA_PATH):

    return Corpus(
        corpus_settings['corpus_name'],
        registry_dir=corpus_settings['registry_dir'],
        lib_dir=corpus_settings.get('lib_dir', None),
        data_dir=data_dir
    )


#################
# CONSTELLATION #
#################
@pytest.mark.now
def test_constellation_init(germaparl, discoursemes):

    corpus = get_corpus(germaparl)

    # init constellation
    topic_query = format_cqp_query(
        discoursemes['topic'],
        p_query=discoursemes['parameters']['p_query'],
        s_query=discoursemes['parameters']['s_query'],
        flags=discoursemes['parameters']['flags_query'],
        escape=discoursemes['parameters']['escape_query']
    )
    topic_dump = corpus.query(
        topic_query,
        context=None,
        context_break=discoursemes['parameters']['s_context']
    )
    const = Constellation(topic_dump)
    # print(const.df)

    assert isinstance(const.df, DataFrame)
    assert len(const.df) == 2789


def test_constellation_add(germaparl, discoursemes):

    corpus = get_corpus(germaparl)

    # init constellation
    topic_query = format_cqp_query(
        discoursemes['topic'],
        p_query=discoursemes['parameters']['p_query'],
        s_query=discoursemes['parameters']['s_query'],
        flags=discoursemes['parameters']['flags_query'],
        escape=discoursemes['parameters']['escape_query']
    )
    topic_dump = corpus.query(
        topic_query,
        context=None,
        context_break=discoursemes['parameters']['s_context']
    )
    const = Constellation(topic_dump)

    # add discourseme
    disc1_query = format_cqp_query(
        discoursemes['disc1'],
        p_query=discoursemes['parameters']['p_query'],
        s_query=discoursemes['parameters']['s_query'],
        flags=discoursemes['parameters']['flags_query'],
        escape=discoursemes['parameters']['escape_query']
    )
    disc1_dump = corpus.query(
        disc1_query,
        context=None,
        context_break=discoursemes['parameters']['s_context']
    )
    const.add_discourseme(disc1_dump)

    assert len(const.df) == 1599
    assert 'topic' in const.discoursemes
    assert len(const.discoursemes) == 2


def test_constellation_add_nodrop(germaparl, discoursemes):

    corpus = get_corpus(germaparl)

    # init constellation
    topic_query = format_cqp_query(
        discoursemes['topic'],
        p_query=discoursemes['parameters']['p_query'],
        s_query=discoursemes['parameters']['s_query'],
        flags=discoursemes['parameters']['flags_query'],
        escape=discoursemes['parameters']['escape_query']
    )
    topic_dump = corpus.query(
        topic_query,
        context=None,
        context_break=discoursemes['parameters']['s_context']
    )
    const = Constellation(topic_dump)
    assert len(const.df) == 2789

    # add discourseme
    disc1_query = format_cqp_query(
        discoursemes['disc1'],
        p_query=discoursemes['parameters']['p_query'],
        s_query=discoursemes['parameters']['s_query'],
        flags=discoursemes['parameters']['flags_query'],
        escape=discoursemes['parameters']['escape_query']
    )
    disc1_dump = corpus.query(
        disc1_query,
        context=None,
        context_break=discoursemes['parameters']['s_context']
    )
    const.add_discourseme(
        disc1_dump,
        drop=False
    )
    assert len(const.df) == 3072


def test_constellation_add2(germaparl, discoursemes):

    corpus = get_corpus(germaparl)

    # init constellation
    topic_query = format_cqp_query(
        discoursemes['topic'],
        p_query=discoursemes['parameters']['p_query'],
        s_query=discoursemes['parameters']['s_query'],
        flags=discoursemes['parameters']['flags_query'],
        escape=discoursemes['parameters']['escape_query']
    )
    topic_dump = corpus.query(
        topic_query,
        context=None,
        context_break=discoursemes['parameters']['s_context']
    )
    const = Constellation(topic_dump)

    # add discourseme 1
    disc1_query = format_cqp_query(
        discoursemes['disc1'],
        p_query=discoursemes['parameters']['p_query'],
        s_query=discoursemes['parameters']['s_query'],
        flags=discoursemes['parameters']['flags_query'],
        escape=discoursemes['parameters']['escape_query']
    )
    disc1_dump = corpus.query(
        disc1_query,
        context=None,
        context_break=discoursemes['parameters']['s_context']
    )
    const.add_discourseme(
        disc1_dump,
        name='disc1'
    )

    # add discourseme 2
    disc2_query = format_cqp_query(
        discoursemes['disc2'],
        p_query=discoursemes['parameters']['p_query'],
        s_query=discoursemes['parameters']['s_query'],
        flags=discoursemes['parameters']['flags_query'],
        escape=discoursemes['parameters']['escape_query']
    )
    disc2_dump = corpus.query(
        disc2_query,
        context=None,
        context_break=discoursemes['parameters']['s_context']
    )
    const.add_discourseme(
        disc2_dump,
        name='disc2'
    )

    assert len(const.df) == 13


########################
# CREATE_CONSTELLATION #
########################
def test_create_constellation(germaparl, discoursemes):

    corpus_name = germaparl['corpus_name']

    # parameters
    parameters = discoursemes.pop('parameters')
    flags = parameters['flags_query']
    escape = parameters['escape_query']
    p_query = parameters['p_query']
    s_query = parameters['s_query']
    s_context = parameters['s_context']
    context = parameters['context']

    # get topic and additional discoursemes
    topic_items = discoursemes.pop('topic')
    topic_discourseme = {
        'topic': topic_items
    }

    # filter
    const = create_constellation(
        corpus_name=corpus_name,
        # discoursemes
        topic_discourseme=topic_discourseme,
        filter_discoursemes=discoursemes,
        additional_discoursemes={},
        # context settings
        s_context=s_context,
        context=context,
        # CWB setttings
        registry_dir=germaparl['registry_dir'],
        data_dir=DATA_PATH,
        # query settings
        p_query=p_query,
        s_query=s_query,
        flags=flags,
        escape=escape
    )

    assert len(const.df) == 12

    # highlight
    const = create_constellation(
        corpus_name=corpus_name,
        # discoursemes
        topic_discourseme=topic_discourseme,
        filter_discoursemes={},
        additional_discoursemes=discoursemes,
        # context settings
        s_context=s_context,
        context=context,
        # CWB setttings
        registry_dir=germaparl['registry_dir'],
        data_dir=DATA_PATH,
        # query settings
        p_query=p_query,
        s_query=s_query,
        flags=flags,
        escape=escape
    )

    assert len(const.df) == 3004


def test_create_textconstellation(germaparl, discoursemes):

    corpus_name = germaparl['corpus_name']

    # parameters
    parameters = discoursemes.pop('parameters')
    flags = parameters['flags_query']
    escape = parameters['escape_query']
    p_query = parameters['p_query']
    s_query = parameters['s_query']
    s_context = parameters['s_context']
    context = parameters['context']

    const = create_constellation(
        corpus_name=corpus_name,
        # discoursemes
        topic_discourseme={},
        filter_discoursemes={},
        additional_discoursemes=discoursemes,
        # context settings
        s_context=s_context,
        context=context,
        # CWB setttings
        registry_dir=germaparl['registry_dir'],
        data_dir=DATA_PATH,
        # query settings
        p_query=p_query,
        s_query=s_query,
        flags=flags,
        escape=escape
    )

    # TODO this yields len(const.df) = 1859 results with Python3.8, but why?
    # assert len(const.df) == 2198


###############
# CONCORDANCE #
###############
def test_constellation_conc(germaparl, discoursemes):

    # parameters
    parameters = discoursemes.pop('parameters')

    # get topic and additional discoursemes
    topic_items = discoursemes.pop('topic')
    topic_discourseme = {
        'topic': topic_items
    }
    discoursemes = discoursemes

    # filter
    const = create_constellation(
        corpus_name=germaparl['corpus_name'],
        # discoursemes
        topic_discourseme=topic_discourseme,
        filter_discoursemes=discoursemes,
        additional_discoursemes={},
        # context settings
        s_context=parameters['s_context'],
        context=parameters['context'],
        # CWB setttings
        registry_dir=germaparl['registry_dir'],
        data_dir=DATA_PATH,
        # query settings
        p_query=parameters['p_query'],
        s_query=parameters['s_query'],
        flags=parameters['flags_query'],
        escape=parameters['escape_query']
    )

    lines = const.concordance(s_show=['text_id'])

    assert len(lines) == 4
    assert isinstance(lines[0], dict)
    assert 'word' in lines[0]
    assert isinstance(lines[0]['word'], list)


def test_constellation_conc_htmlify_meta(germaparl, discoursemes):

    # parameters
    parameters = discoursemes.pop('parameters')

    # get topic and additional discoursemes
    topic_items = discoursemes.pop('topic')
    topic_discourseme = {
        'topic': topic_items
    }
    discoursemes = discoursemes

    # filter
    const = create_constellation(
        corpus_name=germaparl['corpus_name'],
        # discoursemes
        topic_discourseme=topic_discourseme,
        filter_discoursemes=discoursemes,
        additional_discoursemes={},
        # context settings
        s_context=parameters['s_context'],
        context=parameters['context'],
        # CWB setttings
        registry_dir=germaparl['registry_dir'],
        data_dir=DATA_PATH,
        # query settings
        p_query=parameters['p_query'],
        s_query=parameters['s_query'],
        flags=parameters['flags_query'],
        escape=parameters['escape_query']
    )

    lines = const.concordance(s_show=['text_id'], htmlify_meta=True)

    # assert len(lines) == 4  error on github: actually 100 lines, but why?
    assert isinstance(lines[0], dict)
    assert 'word' in lines[0]
    assert isinstance(lines[0]['word'], list)
    assert isinstance(lines[0]['meta'], str)


###############
# COLLOCATION #
###############
def test_constellation_collocates(germaparl):

    topic_name = 'topic'
    topic_items = ['CDU', 'CSU']
    p_query = 'lemma'
    s_query = None
    flags_query = '%cd'
    flags_show = ''
    min_freq = 2
    s_context = 's'
    context = 20
    additional_discoursemes = {}
    windows = [3, 5, 7]
    cqp_bin = 'cqp'
    lib_dir = None
    p_show = ['lemma']
    ams = None
    cut_off = 200
    min_freq = 2
    order = 'log_likelihood'
    escape = True
    match_strategy = 'longest'

    # preprocess parameters
    s_query = s_context if s_query is None else s_query
    topic_name = 'topic'

    # create constellation
    const = create_constellation(
        corpus_name=germaparl['corpus_name'],
        # discoursemes
        topic_discourseme={topic_name: topic_items},
        filter_discoursemes={},
        additional_discoursemes=additional_discoursemes,
        # context settings
        s_context=s_context,
        context=context,
        # CWB setttings
        match_strategy=match_strategy,
        lib_dir=lib_dir,
        cqp_bin=cqp_bin,
        registry_dir=germaparl['registry_dir'],
        data_dir=DATA_PATH,
        # query settings
        p_query=p_query,
        s_query=s_query,
        flags=flags_query,
        escape=escape
    )

    collocates = const.collocates(windows=windows, p_show=p_show,
                                  flags=flags_show, ams=ams,
                                  min_freq=min_freq, order=order,
                                  cut_off=cut_off)

    assert len(collocates) == 3


def test_constellation_coll(germaparl, discoursemes):

    corpus = get_corpus(germaparl)

    # init constellation
    topic_query = format_cqp_query(
        discoursemes['topic'],
        p_query=discoursemes['parameters']['p_query'],
        s_query=discoursemes['parameters']['s_query'],
        flags=discoursemes['parameters']['flags_query'],
        escape=discoursemes['parameters']['escape_query']
    )
    topic_dump = corpus.query(
        topic_query,
        context=None,
        context_break=discoursemes['parameters']['s_context']
    )
    const = Constellation(topic_dump)

    # add discourseme 1
    disc1_query = format_cqp_query(
        discoursemes['disc1'],
        p_query=discoursemes['parameters']['p_query'],
        s_query=discoursemes['parameters']['s_query'],
        flags=discoursemes['parameters']['flags_query'],
        escape=discoursemes['parameters']['escape_query']
    )
    disc1_dump = corpus.query(
        disc1_query,
        context=None,
        context_break=discoursemes['parameters']['s_context']
    )
    const.add_discourseme(
        disc1_dump,
        name='disc1'
    )

    # add discourseme 2
    disc2_query = format_cqp_query(
        discoursemes['disc2'],
        p_query=discoursemes['parameters']['p_query'],
        s_query=discoursemes['parameters']['s_query'],
        flags=discoursemes['parameters']['flags_query'],
        escape=discoursemes['parameters']['escape_query']
    )
    disc2_dump = corpus.query(
        disc2_query,
        context=None,
        context_break=discoursemes['parameters']['s_context']
    )
    const.add_discourseme(
        disc2_dump,
        name='disc2'
    )

    dfs = const.collocates(windows=list(range(1, 21)))
    assert len(dfs) == 20
    assert len(dfs[1]) == 2
    # if approximate, i.e. nodes are not removed:
    # assert len(dfs[20]) == 8
    assert len(dfs[20]) == 5


##########################
# TEXTUAL CONSTELLATIONS #
##########################
def test_textual_constellation(germaparl, discoursemes):

    corpus = get_corpus(germaparl)

    # init constellation
    topic_query = format_cqp_query(
        discoursemes['topic'],
        p_query=discoursemes['parameters']['p_query'],
        s_query=discoursemes['parameters']['s_query'],
        flags=discoursemes['parameters']['flags_query'],
        escape=discoursemes['parameters']['escape_query']
    )
    topic_dump = corpus.query(
        topic_query,
        context=None,
        context_break=discoursemes['parameters']['s_context']
    )
    const = TextConstellation(
        topic_dump,
        s_context=discoursemes['parameters']['s_context']
    )
    assert len(const.df) == 624
    assert 'topic' in const.df.columns

    # print(const.df)


def test_textual_constellation_add(germaparl, discoursemes):

    corpus = get_corpus(germaparl)

    # init constellation
    topic_query = format_cqp_query(
        discoursemes['topic'],
        p_query=discoursemes['parameters']['p_query'],
        s_query=discoursemes['parameters']['s_query'],
        flags=discoursemes['parameters']['flags_query'],
        escape=discoursemes['parameters']['escape_query']
    )
    topic_dump = corpus.query(
        topic_query,
        context=None,
        context_break=discoursemes['parameters']['s_context']
    )
    const = TextConstellation(
        topic_dump,
        s_context=discoursemes['parameters']['s_context']
    )

    # add discourseme
    disc1_query = format_cqp_query(
        discoursemes['disc1'],
        p_query=discoursemes['parameters']['p_query'],
        s_query=discoursemes['parameters']['s_query'],
        flags=discoursemes['parameters']['flags_query'],
        escape=discoursemes['parameters']['escape_query']
    )
    disc1_dump = corpus.query(
        disc1_query,
        context=None,
        context_break=discoursemes['parameters']['s_context']
    )
    const.add_discourseme(
        disc1_dump
    )

    assert len(const.df) == 2156
    assert 'discourseme' in const.df.columns


def test_textual_constellation_association(germaparl, discoursemes):

    # parameters
    parameters = discoursemes.pop('parameters')
    flags = parameters['flags_query']
    escape = parameters['escape_query']
    p_query = parameters['p_query']
    s_query = parameters['s_query']
    s_context = parameters['s_context']
    context = parameters['context']

    const = create_constellation(
        corpus_name=germaparl['corpus_name'],
        # discoursemes
        topic_discourseme={},
        filter_discoursemes=discoursemes,
        additional_discoursemes={},
        # context settings
        s_context=s_context,
        context=context,
        # CWB setttings
        registry_dir=germaparl['registry_dir'],
        data_dir=DATA_PATH,
        # query settings
        p_query=p_query,
        s_query=s_query,
        flags=flags,
        escape=escape
    )

    assoc = const.associations()
    assert len(assoc) == 6
    assert 'candidate' in assoc.columns


def test_textual_constellation_association_approximate(germaparl, discoursemes):

    # parameters
    parameters = discoursemes.pop('parameters')
    flags = parameters['flags_query']
    escape = parameters['escape_query']
    p_query = parameters['p_query']
    s_query = parameters['s_query']
    s_context = parameters['s_context']
    context = parameters['context']

    const = create_constellation(
        corpus_name=germaparl['corpus_name'],
        # discoursemes
        topic_discourseme={},
        filter_discoursemes=discoursemes,
        additional_discoursemes={},
        # context settings
        s_context=s_context,
        context=context,
        approximate=True,
        # CWB setttings
        registry_dir=germaparl['registry_dir'],
        data_dir=DATA_PATH,
        # query settings
        p_query=p_query,
        s_query=s_query,
        flags=flags,
        escape=escape
    )

    assoc = const.associations()
    assert len(assoc) == 6
    assert 'candidate' in assoc.columns


def test_textual_constellation_association_empty(germaparl, discoursemes):

    # parameters
    parameters = discoursemes.pop('parameters')
    flags = parameters['flags_query']
    escape = parameters['escape_query']
    p_query = parameters['p_query']
    s_query = parameters['s_query']
    s_context = parameters['s_context']
    context = parameters['context']
    discoursemes2 = discoursemes.copy()
    discoursemes2['fail'] = ["fail"]
    discoursemes2['fail2'] = ["fail2"]

    const = create_constellation(
        corpus_name=germaparl['corpus_name'],
        # discoursemes
        topic_discourseme={},
        filter_discoursemes=discoursemes,
        additional_discoursemes={},
        # context settings
        s_context=s_context,
        context=context,
        approximate=True,
        # CWB setttings
        registry_dir=germaparl['registry_dir'],
        data_dir=DATA_PATH,
        # query settings
        p_query=p_query,
        s_query=s_query,
        flags=flags,
        escape=escape
    )

    assoc = const.associations()
    assert len(assoc) == 6
    assert 'candidate' in assoc.columns


def test_textual_constellation_concordance(germaparl, discoursemes):

    # parameters
    parameters = discoursemes.pop('parameters')
    flags = parameters['flags_query']
    escape = parameters['escape_query']
    p_query = parameters['p_query']
    s_query = parameters['s_query']
    s_context = parameters['s_context']
    context = parameters['context']

    # create constellation
    const = create_constellation(
        corpus_name=germaparl['corpus_name'],
        # discoursemes
        topic_discourseme={},
        filter_discoursemes=discoursemes,
        additional_discoursemes={},
        # context settings
        s_context=s_context,
        context=context,
        # CWB setttings
        registry_dir=germaparl['registry_dir'],
        data_dir=DATA_PATH,
        # query settings
        p_query=p_query,
        s_query=s_query,
        flags=flags,
        escape=escape
    )

    # retrieve lines
    lines = const.concordance(cut_off=None)

    assert len(lines) == 2198


def test_textual_constellation_breakdown(germaparl, discoursemes):

    # parameters
    parameters = discoursemes.pop('parameters')
    flags = parameters['flags_query']
    escape = parameters['escape_query']
    p_query = parameters['p_query']
    s_query = parameters['s_query']
    s_context = parameters['s_context']
    context = parameters['context']

    # create constellation
    const = create_constellation(
        corpus_name=germaparl['corpus_name'],
        # discoursemes
        topic_discourseme={},
        filter_discoursemes=discoursemes,
        additional_discoursemes={},
        # context settings
        s_context=s_context,
        context=context,
        # CWB setttings
        registry_dir=germaparl['registry_dir'],
        data_dir=DATA_PATH,
        # query settings
        p_query=p_query,
        s_query=s_query,
        flags=flags,
        escape=escape
    )

    assert len(const.breakdown()) == 5



import os
from random import randint

from pandas import DataFrame

from ccc.cache import Cache, generate_idx, generate_library_idx

from .conftest import DATA_PATH


def test_set_get_empty():

    parameters = {'query': "test", 's_break': "test"}
    cache = Cache()
    s = generate_idx(parameters)
    assert cache.get(s) is None
    cache.set(s, parameters)
    assert cache.get(s) is None


def test_set_get():

    cache = Cache(os.path.join(DATA_PATH, 'test-cache'))
    parameters = {'query': "test", 's_break': "test", 'key': randint(0, 9 * 10 ^ 9)}
    dump = DataFrame()
    r = cache.get(parameters.values())
    assert r is None

    cache.set(parameters, dump)
    r = cache.get(parameters)
    assert r.empty

    cache.set('testtest', dump)
    r = cache.get('testtest')
    assert r.empty


def test_generate_library_idx(germaparl):
    assert isinstance(generate_library_idx(germaparl['lib_dir']), str)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""cache.py

simple caching using shelve/pickle.

"""
import logging
import os
import shelve
from glob import glob
from hashlib import sha256

logger = logging.getLogger(__name__)


def generate_idx(identifiers, prefix='', length=10):
    """generate an ID from an iterable

    """
    string = ''.join([str(idx) for idx in identifiers])
    identifier = sha256(str(string).encode()).hexdigest()
    return prefix + identifier[:length]


def generate_library_idx(lib_dir, prefix='', length=10):
    """generate an ID for a library,
    i.e. all files ending on '.txt' in "wordlists" and "macros" of lib_dir

    """
    wordlists = glob(os.path.join(lib_dir, 'wordlists', '*.txt'))
    macros = glob(os.path.join(lib_dir, 'macros', '*.txt'))
    paths = macros + wordlists
    identifiers = [
        generate_idx(open(p, 'rt', encoding='utf-8')) for p in sorted(paths)
    ]
    return generate_idx(identifiers, prefix, length)


class Cache:

    def __init__(self, path=None):

        self.path = path

        if path:
            directory = os.path.dirname(path)
            os.makedirs(directory, exist_ok=True)

    def delete(self, identifier):

        if self.path is None:
            logger.info('no cache path')
            return

        if isinstance(identifier, str):
            key = identifier
        else:
            key = generate_idx(identifier)

        with shelve.open(self.path) as shelf:
            if key in shelf.keys():
                logger.info(f'deleting object "{key}" from cache')
                del shelf[key]

    def get(self, identifier):

        if self.path is None:
            logger.info('no cache path')
            return

        if isinstance(identifier, str):
            key = identifier
        else:
            key = generate_idx(identifier)

        with shelve.open(self.path) as shelf:
            if key in shelf.keys():
                logger.info(f'retrieving object "{key}" from cache')
                return shelf[key]

    def set(self, identifier, value):

        if self.path is None:
            logger.error('no cache path')
            return

        if isinstance(identifier, str):
            key = identifier
        else:
            key = generate_idx(identifier)

        with shelve.open(self.path) as shelf:
            logger.info(f'saving object "{key}" to cache')
            shelf[key] = value



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""cqp.py

access to CQP

original version by Joerg Asmussen (2008)
current version by Philipp Heinrich (2023)

"""
import logging
import os
import random
import re
import select
import subprocess
import sys
import threading
import time
from glob import glob
from io import StringIO
from tempfile import NamedTemporaryFile

# requirements
from pandas import DataFrame, read_csv

logger = logging.getLogger(__name__)


# GLOBAL CONSTANTS OF MODULE:
CPROGRESSCONTROLCYCLE = 5   # secs between each progress control cycle
CMAXREQUESTPROCTIME = 900   # max secs for processing a user request


# ERROR MESSAGE TYPES:
class ErrCQP:
    """Handle CQP error message."""
    def __init__(self, msg):
        """Class constructor."""
        self.msg = msg.rstrip()


class ErrKilled:
    """Handle errors when process is killed."""
    def __init__(self, msg):
        """Class constructor."""
        self.msg = msg.rstrip()


# ACTUAL INTERFACE:
class CQP:
    """Wrapper for CQP."""

    def _progressController(self):
        """Control the progress.

        CREATED: 2008-02

        This method is run as a thread.
        At certain intervals (CPROGRESSCONTROLCYCLE), it controls how long the
        CQP process of the current user has spent on processing this user's
        latest CQP command. If this time exceeds a certain maximum
        (CMAXREQUESTPROCTIME), this method kills the CQP process.
        """
        self.runController = True
        while self.runController:
            time.sleep(CPROGRESSCONTROLCYCLE)
            if self.execStart is not None:
                if time.time() - self.execStart > CMAXREQUESTPROCTIME * self.maxProcCycles:
                    logger.error(f"progress controller identified blocking cqp process id {self.CQP_process.pid}")
                    os.popen("kill -9 " + str(self.CQP_process.pid))
                    self.CQPrunning = False
                    self.execStart = None
                    logger.error("-- CQP process killed!")
                    break

    def __init__(self, binary="cqp", options='-c', print_version=False):
        """Class constructor."""
        self.execStart = time.time()
        self.maxProcCycles = 1.0

        # start CQP as a child process of this wrapper
        if binary is None:
            logger.error("path to CQP binaries undefined")
            sys.exit(1)
        self.CQP_process = subprocess.Popen(binary + ' ' + options,
                                            shell=True,
                                            stdin=subprocess.PIPE,
                                            stdout=subprocess.PIPE,
                                            stderr=subprocess.PIPE,
                                            universal_newlines=True,
                                            close_fds=True,
                                            preexec_fn=os.setsid)
        self.CQPrunning = True

        # init progress controller
        progressthread = threading.Thread(target=self._progressController, args=())
        progressthread.daemon = True
        progressthread.start()

        # "cqp -c" should print version on startup:
        version_string = self.CQP_process.stdout.readline()
        version_string = version_string.rstrip()  # Equivalent to Perl's chomp
        self.CQP_process.stdout.flush()
        if print_version:
            print(version_string)
        logger.debug("CQP " + "-" * 43 + " started")
        version_regexp = re.compile(
            r'^CQP\s+(?:\w+\s+)*([0-9]+)\.([0-9]+)(?:\.b?([0-9]+))?(?:\s+(.*))?$'
        )
        match = version_regexp.match(version_string)
        if not match:
            logger.error("CQP backend startup failed")
            sys.exit(1)
        self.major_version = int(match.group(1))
        self.minor_version = int(match.group(2))
        self.beta_version = int(match.group(3))
        self.compile_date = match.group(4)

        # We need cqp-2.2.b41 or newer (for query lock):
        if not (self.major_version >= 3 or (self.major_version == 2 and
                                            self.minor_version == 2 and
                                            self.beta_version >= 41)):
            logger.error("CQP version too old: " + version_string)
            sys.exit(1)

        # Error handling:
        self.error_handler = None
        self.status = 'ok'
        self.error_message = ''  # we store compound error messages as a STRING
        self.errpipe = self.CQP_process.stderr.fileno()

        # CQP defaults:
        self.Exec('set PrettyPrint off')
        self.execStart = None

    def SetProcCycles(self, procCycles):
        """Set procCycles."""
        print(f"    Setting procCycles to {procCycles}")
        self.maxProcCycles = procCycles
        return int(self.maxProcCycles * CMAXREQUESTPROCTIME)

    def __del__(self):
        """Stop running CQP instance."""
        if self.CQPrunning:
            logger.debug("Shutting down CQP backend ...")
            self.runController = False
            self.execStart = time.time()
            self.CQP_process.stdin.write('exit;')
            self.CQP_process.stdin.flush()
            del self.CQP_process
            self.CQPrunning = False
            self.execStart = None
            logger.debug("... -- CQP object deleted.")

    def Exec(self, cmd):
        """Execute CQP command.

        The method takes as input a command string and sends it
        to the CQP child process
        """
        self.execStart = time.time()
        self.status = 'ok'
        cmd = cmd.rstrip()  # Equivalent to Perl's 'chomp'
        cmd = re.sub(r';\s*$', r'', cmd)
        logger.debug("CQP << " + cmd + ";")
        try:
            self.CQP_process.stdin.write(cmd + '; .EOL.;\n')
        except IOError:
            return None
        # In CQP.pm lines are appended to a list @result.
        # This implementation prefers a string structure instead
        # because output from this module is meant to be transferred
        # accross a server connection. To enable the development of
        # client modules written in any language, the server only emits
        # strings which then are to be structured by the client module.
        # The server does not emit pickled data according to some
        # language dependent protocol.
        self.CQP_process.stdin.flush()
        result = []
        while self.CQPrunning:
            ln = self.CQP_process.stdout.readline()
            ln = ln.strip()  # strip off whitespace from start and end of line
            if re.match(r'-::-EOL-::-', ln):
                logger.debug("CQP " + "-" * 40 + " terminated")
                break
            logger.debug("CQP >> " + ln)
            if ln != '':
                result.append(ln)
            self.CQP_process.stdout.flush()
        self.Checkerr()
        self.execStart = None
        result = '\n'.join(result)
        result = result.rstrip()  # strip off whitespace from EOL (\n)
        return result

    def Query(self, query):
        """Execute query in safe mode (query lock)."""
        result = []
        key = str(random.randint(1, 1000000))
        errormsg = ''  # collect CQP error messages AS STRING
        ok = True      # check if any error occurs
        self.Exec('set QueryLock ' + key)  # enter query lock mode
        if self.status != 'ok':
            errormsg = errormsg + self.error_message
            ok = False
        result = self.Exec(query)
        if self.status != 'ok':
            errormsg = errormsg + self.error_message.decode('utf-8')
            ok = ok and False
        self.Exec('unlock ' + key)  # unlock with random key
        if self.status != 'ok':
            errormsg = errormsg + self.error_message
            ok = ok and False
        # Set error status & error message:
        if ok:
            self.status = 'ok'
        else:
            self.status = 'error'
            self.error_message = errormsg
        return result

    def Dump(self, subcorpus='Last', first=None, last=None):
        """Dump named query result into table of corpus positions."""

        # check first and last
        if first is None and last is None:
            # actual dump
            result = self.Exec('dump ' + subcorpus + ";")

        elif ((not isinstance(first, int) and first is not None) or
              (not isinstance(last, int) and last is not None)):
            logger.error("Invalid value for first (" + str(first) +
                         ") or last (" + str(last) + ") line in Dump() method")
            sys.exit(1)
        elif isinstance(first, int) and isinstance(last, int):
            if first > last:
                logger.error("Invalid value for first line (first = " +
                             str(first) + " > last = " + str(last) +
                             ") in Dump() method")
                sys.exit(1)
        else:
            if first is not None and last is None:
                last = first
            elif last is not None and first is None:
                first = last
            # actual dump
            result = self.Exec(
                'dump ' + subcorpus + " " + str(first) + " " + str(last) + ";"
            )

        # convert to pandas dataframe
        df = read_csv(StringIO(result),
                      sep="\t", header=None, index_col=[0, 1],
                      names=["match", "matchend", "target", "keyword"])
        df = df.astype(int)

        return df

    def Undump(self, subcorpus="Last", df=DataFrame()):
        """Undump named query result from table of corpus positions."""

        columns = ['match', 'matchend']
        wth = ''
        if 'target' in df.columns:
            wth = 'with target '
            columns = columns + ['target']
        if 'keyword' in df.columns:
            wth = 'with target keyword '
            columns = columns + ['keyword']

        with NamedTemporaryFile(mode='wt') as f:
            f.write(str(len(df)) + "\n")
            df.reset_index().to_csv(f, mode="a", sep="\t", columns=columns, header=None, index=False)
            f.seek(0)
            f.flush()
            self.Exec("undump " + subcorpus + " " + wth + '< "' + f.name + '";')

    def Group(self, subcorpus='Last',
              spec1='match.word', spec2='', cutoff='1'):
        """Compute frequency distribution over attribute values.

        (single values or pairs) using group command.

        Note that the arguments are specified in the logical order,
        in contrast to "group"
        """
        spec2_regexp = re.compile(r'^[0-9]+$')
        if spec2_regexp.match(spec2):
            cutoff = spec2
            spec2 = ''
        spec_regexp = re.compile(
          r'^(match|matchend|target[0-9]?|keyword)\.([A-Za-z0-9_-]+)$'
        )
        match = re.match(spec_regexp, spec1)
        if not match:
            logger.error("Invalid key '" + spec1 + "' in Group() method")
            sys.exit(1)
        spec1 = match.group(1) + ' ' + match.group(2)
        if spec2 != '':
            match = re.match(spec_regexp, spec2)
            if not match:
                logger.error("Invalid key '" + spec2 + "' in Group() method")
                sys.exit(1)
            spec2 = match.group(1) + ' ' + match.group(2)
            cmd = 'group ' + subcorpus + ' ' + spec2 + ' by ' + spec1 + \
                  ' cut ' + cutoff
        else:
            cmd = 'group ' + subcorpus + ' ' + spec1 + ' cut ' + cutoff
        result = self.Exec(cmd)
        return result

    def Count(self, subcorpus='Last', sort_clause=None, cutoff=1):
        """Compute frequency distribution for match strings.

        Based on sort clause.
        """
        if sort_clause is None:
            logger.error("Parameter 'sort_clause' undefined in Count() method")
            sys.exit(1)
        return self.Exec('count ' + subcorpus +
                         ' by ' + sort_clause +
                         ' cut ' + str(cutoff))

    def Checkerr(self):
        """Check CQP's stderr stream for error messages.

        (returns true if there was an error).
        OBS! In CQP.pm the error_message is stored in a list,
        In PyCQP_interface.pm we use a string (which better can be sent
        accross the server line).
        """
        ready = select.select([self.errpipe], [], [], 0)
        if self.errpipe in ready[0]:
            # We've got something on stderr -> an error must have occurred:
            self.status = 'error'
            self.error_message = self.Readerr()
        return not self.Ok()

    def Readerr(self):
        """Read all available lines from CQP's stderr stream."""
        return os.read(self.errpipe, 16384)

    def Status(self):
        """Read the CQP object's (error) status."""
        return self.status

    def Ok(self):
        """Simplified interface for checking for CQP errors."""
        if self.CQPrunning:
            return (self.Status() == 'ok')
        else:
            return False

    def Error_message(self):
        """Return the CQP error message."""
        if self.CQPrunning:
            return ErrCQP(self.error_message)
        else:
            msgKilled = '**** CQP KILLED ***\n\
            CQP COULD NOT PROCESS YOUR REQUEST\n'
            return ErrKilled(msgKilled + self.error_message)

    def Error(self, msg):
        """Processe/output error messages.

        (optionally run through user-defined error handler)
        """
        if self.error_handler is not None:
            self.error_handler(msg)
        else:
            logger.error(msg)

    def Set_error_handler(self, handler=None):
        """Set user-defined error handler."""
        self.error_handler = handler

    #########################
    # SOME ALIASES FOR NQRs #
    #########################
    def nqr_from_query(self, query, name='Last',
                       match_strategy='longest', return_dump=True,
                       propagate_error=False):
        """Defines NQR from query, optionally returns dump.

        :param str query: valid CQP query
        :param str name: name for NQR
        :param str match_strategy: CQP matching strategy
        :param bool return_dump: whether to return the dump

        :return: df_dump
        :rtype: DataFrame

        """
        name = 'Last' if name is None else name

        logger.info(f'defining NQR "{name}" from query: {query}')
        self.Query(f'{name}={query};')

        if not self.Ok():
            logger.error(f'{self.error_message}')
            return self.error_message if propagate_error else DataFrame()

        size = int(self.Exec(f"size {name}"))
        if size == 0:
            logger.info(f'no results for query: {query}')
            return DataFrame() if return_dump else None

        if return_dump:
            logger.info('dumping result')
            df_dump = self.Dump(name)
            return df_dump

    def nqr_from_dump(self, df_dump, name='Last'):
        """Alias for Undump. Defines NQR from given dump.

        :param DataFrame df_dump: DataFrame indexed by (match, matchend)
                                  with optional columns 'target' and 'keyword'
        :param str name: name for NQR

        """
        logger.info(f'defining NQR "{name}" from dump with {len(df_dump)} matches')
        self.Undump(name, df_dump)
        if not self.Ok():
            logger.error('invalid dump')

    def nqr_activate(self, corpus_name, name=None):
        """Activate NQR or whole corpus.

        :param str corpus_name: name of the corpus
        :param str name: name of NQR

        """
        if name is not None:
            logger.info(f'activating NQR "{corpus_name}:{name}"')
            self.Exec(name)

        else:
            logger.info(f'activating corpus "{corpus_name}"')
            self.Exec(corpus_name)

        if not self.Ok():
            logger.error('invalid corpus or NQR')

    def nqr_save(self, corpus_name, name='Last'):
        """Save NQR to disk.

        :param str corpus_name: name of the corpus
        :param str name: nqr to save

        """
        logger.info(f'saving NQR "{corpus_name}:{name}" to disk')
        self.Exec(f"save {name};")
        if not self.Ok():
            logger.error('invalid corpus or NQR')


def start_cqp(cqp_bin, registry_dir,
              data_dir=None, corpus_name=None,
              lib_dir=None, subcorpus_name=None):
    """Start CQP process, activate (sub-)corpus, set directory paths, and
    read library (macros and wordlists).

    :param str cqp_bin: /path/to/cqp-binary
    :param str registry_dir: /path/to/cwb/registry/
    :param str data_dir: /path/to/data/and/cache/
    :param str corpus_name: name of corpus in CWB registry
    :param str lib_dir: /path/to/macros/and/wordlists/
    :param str subcorpus_name: name of subcorpus (NQR)

    :return: CQP process
    :rtype: CQP

    """

    cqp = CQP(
        binary=cqp_bin,
        options='-c -r ' + registry_dir
    )

    if data_dir is not None:
        cqp.Exec(f'set DataDirectory "{data_dir}"')

    if lib_dir is not None:

        # wordlists
        wordlists = glob(os.path.join(lib_dir, 'wordlists', '*.txt'))
        for wordlist in wordlists:
            name = wordlist.split('/')[-1].split('.')[0]
            abs_path = os.path.abspath(wordlist)
            cqp_exec = f'define ${name} < "{abs_path}";'
            cqp.Exec(cqp_exec)
        # macros
        macros = glob(os.path.join(lib_dir, 'macros', '*.txt'))
        for macro in macros:
            abs_path = os.path.abspath(macro)
            cqp_exec = f'define macro < "{abs_path}";'
            cqp.Exec(cqp_exec)
        # for wordlists defined in macros, it is necessary to execute the macro once
        macros = cqp.Exec("show macro;").split("\n")
        for macro in macros:
            # NB: this yields !cqp.Ok() if macro is not zero-valent
            cqp.Exec(macro.split("(")[0] + "();")

    # initialize corpus after macro definition, so execution of macro doesn't spend time
    if corpus_name is not None:
        cqp.Exec(corpus_name)
    if subcorpus_name is not None:
        cqp.Exec(subcorpus_name)

    if not cqp.Ok():
        raise NotImplementedError(cqp.error_message.decode())

    return cqp



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""concordances.py

definition of the Concordance class.

"""
import itertools
import logging
from random import sample, seed

# requirements
from pandas import DataFrame

# part of module
from .utils import cqp_escape, decode, _node2cotext

logger = logging.getLogger(__name__)


class Concordance:
    """ concordancing """

    def __init__(self, corpus, df_dump):

        # bind corpus
        self.corpus = corpus

        # make sure there's context
        if 'context' not in df_dump.columns:
            df_dump['context'] = df_dump.index.get_level_values('match')
        if 'contextend' not in df_dump.columns:
            df_dump['contextend'] = df_dump.index.get_level_values('matchend')

        # bind dump and check size
        self.df_dump = df_dump
        self.size = len(df_dump)
        if self.size == 0:
            logger.warning('empty dump')

        # init anchor points
        anchors = [i for i in range(10) if i in df_dump.columns]
        self.anchors = anchors

    def simple(self, df, p_show=['word'], start='context', end='contextend'):
        """Retrieve concordance lines of provided df in 'simple'
        formatting, i.e. one column for each $p in p_show.

        """
        for p_att in p_show:
            df = self.corpus.dump2patt(
                df, p_att=p_att, start=start, end=end
            )

        return df[p_show]

    def kwic(self, df, p_show=['word']):
        """Retrieve concordance lines of provided df in 'kwic' formatting. The
        resulting dataframe has 3 times len(p_show) columns, namely
        for each $p in p_show:
        - left_$p: context .. match - 1
        - node_$p: match .. matchend
        - right_$p: matchend + 1 .. contextend

        """

        df = df.reset_index()
        df['leftend'] = df['match'] - 1
        df['rightstart'] = df['matchend'] + 1
        df = df.set_index(['match', 'matchend'])

        for p in p_show:
            df['left' + "_" + p] = self.corpus.dump2patt(
                df, p, start='context', end='leftend'
            )[p]
            df['node' + "_" + p] = self.corpus.dump2patt(
                df, p, start='match', end='matchend'
            )[p]
            df['right' + "_" + p] = self.corpus.dump2patt(
                df, p, start='rightstart', end='contextend'
            )[p]

        return df[["_".join(combo) for combo in itertools.product(
            ['left', 'node', 'right'], p_show
        )]]

    def slots(self, df, p_show=['word'], slots=None):
        """Retrieve concordance lines of provided df in 'slot' formatting.
        Slots are singletons or pairs of self.anchors.  If no slots
        are provided, all anchors and match and matchend are slots.

        If slots is a list (of singletons and / or pairs), it will be
        translated into a dictionary with keys of the form
        "$slot[0]..$slot[1]".

        For each single slot $s and each p-attribute $p in p_show,
        $slot_$p is retrieved.  For pairs, $slot[0]..$slot[1]_$p is
        retrieved.

        If $slot[0] or $slot[1] is not defined in a row (hence the
        entry is -1), the pair is treated as a singleton.  If both
        positions are not defined, an empty string is returned.

        """

        # init output with simple view
        df_lines = self.simple(df, p_show)

        # get slots
        if slots is None:
            slots = self.anchors + [('match', 'matchend')]

        # translate list into dict
        if isinstance(slots, list):
            slots_dict = dict()
            for slot in slots:
                if isinstance(slot, (int, str)):
                    key = slot
                else:
                    key = "..".join([str(s) for s in slot])
                slots_dict[key] = slot
            slots = slots_dict

        # add slots
        for key in slots.keys():
            # allow definition of slots as singletons
            if isinstance(slots[key], (int, str)):
                start = end = slots[key]
            # allow definition via lists of length 1
            elif isinstance(slots[key], list) and len(slots[key]) == 1:
                start = end = slots[key][0]
            # proper definition as pair / list of two
            else:
                start, end = slots[key]

            # retrieve concordance lines
            df_lines[[str(key) + "_" + p for p in p_show]] = self.simple(
                df, p_show=p_show, start=start, end=end
            )[p_show]

        return df_lines

    def _export(self, index, row, p_show, form='dict'):
        """Translates one row of a df_dump into a dictionary or a DataFrame.

        :return: [cpos, offset, anchors] + p_show
        :rtype: dict or DataFrame

        """

        # pack all values into a dictionary
        row = dict(row)
        match, matchend = index
        row['match'] = match
        row['matchend'] = matchend

        # create cotext
        cotext = _node2cotext(row['match'], row['matchend'], row['context'], row['contextend'])

        # lexicalize positions
        attribute_lists = zip(
            *list(map(lambda x: self.corpus.cpos2patts(x, p_show), cotext['cpos_list']))
        )

        # init output dictionary
        d = {
            'cpos': cotext['cpos_list'],
            'offset': cotext['offset_list']
        }
        # add attributes
        for p, a in zip(p_show, attribute_lists):
            d[p] = list(a)

        if form == 'dict':
            # add match as identifier
            d['match'] = match
            # process anchors
            d['anchors'] = dict()
            for a in self.anchors:
                d['anchors'][a] = row[a]
            return d

        elif form == 'dataframe':
            # convert to dataframe and sort columns
            df = DataFrame.from_records(d).set_index('cpos')
            df = df[['offset'] + p_show]
            # process anchors
            for a in self.anchors:
                df[a] = False
                if row[a] in df.index:
                    df.loc[row[a], a] = True
            return df

    def dict(self, df, p_show=['word']):
        """Retrieve concordance lines of provided df in 'dict' formatting.
        Primary export format.

        Each entry in column 'dict' has the following keys:
        # identifier
        - match <int>: single integer as ID
        # aligned token-lists
        - cpos <list>: corpus position
        - offset <list>: offset to match..matchend
        - $p for $p in p_show <list>: surface of p-attribute, each a <list>
        # roles
        - anchors <dict>: {anchor: cpos}

        """

        df['dict'] = df.apply(
            lambda row: self._export(row.name, row, p_show, form='dict'),
            axis=1
        )

        return df

    def dataframe(self, df, p_show=['word']):
        """Retrieve concordance lines of provided df_dict in 'dataframe'
        formatting.  df_dict must contain column 'dict' (return value
        of self.dict()).

        Each concordance line is a DataFrame with the cpos of each
        token as index:

        == (cpos), offset, p_show[0], p_show[1], ... ==

        """

        df['dataframe'] = df.apply(
            lambda row: self._export(row.name, row, p_show, form='dataframe'),
            axis=1
        )

        return df

    def lines(self, form='simple', p_show=['word'], s_show=[],
              order='first', cut_off=100, matches=None, slots=None,
              cwb_ids=False):
        """Retrieve concordance lines from self.df_dump.  Central entry point
        for all methods.  Functionality includes:

        1. simple / kwic / dict / slots / dataframe format
        2. selection of p-attributes and s-attributes
        3. selection of matches using cut-off and order or a
           pre-defined list of matches
        4. definition of slots via anchor points (only for form='slot')

        Return value of all formats is a DataFrame indexed by (match,
        matchend) and requested s-attributes of the match:

        == (match, matchend), s_show[0], ... ==

        Further columns depend on output format, see the respective
        format implementations.

        :param str form: simple / kwic / dict / slots / dataframe
        :param list p_show: p-attributes to retrieve
        :param list s_show: s-attributes (at match position)
        :param str order: first / last / random
        :param int cut_off: how many lines to retrieve
        :param list matches: specific matches to retrieve
        :param dict slots: definition of slots via anchor points

        :return: concordance lines
        :rtype: DataFrame

        """

        # check parameters
        if len(self.df_dump) == 0:
            logger.error("no concordance lines to show")
            return DataFrame()
        if form not in ['simple', 'kwic', 'dict', 'dataframe', 'slots']:
            logger.error('format not implemented, using "simple" format')
            form = 'simple'
        if isinstance(p_show, str):
            p_show = [p_show]
        if isinstance(s_show, str):
            p_show = [s_show]

        # select appropriate subset of matches
        logger.info('lines: selecting matches')
        if matches is None:
            # select all matches
            matches = list(self.df_dump.index.droplevel('matchend'))
        # pre-process cut_off if necessary
        if (cut_off is None) or (len(matches) < cut_off):
            cut_off = len(matches)
        # order
        if isinstance(order, int):
            seed(order)
            order = 'random'

        if order == 'random':
            matches = sample(list(matches), cut_off)
        elif order == 'first':
            matches = sorted(list(matches))[:cut_off]
        elif order == 'last':
            matches = sorted(list(matches))[-cut_off:]
        elif order == 'asis':
            pass
        else:
            logger.error('order not implemented, using "random" order')
            order = 'random'
            matches = sample(list(matches), cut_off)
        logger.info(f"lines: retrieving {len(matches)} concordance line(s)")
        df = self.df_dump.loc[list(matches), :]

        # retrieve p-attributes in respective formatting
        if len(p_show) > 0:
            if form == 'simple':
                df = self.simple(df, p_show)
            elif form == 'kwic':
                df = self.kwic(df, p_show)
            elif form == 'slots':
                df = self.slots(df, p_show, slots)
            elif form == 'dict':
                df = self.dict(df, p_show)
            elif form == 'dataframe':
                df = self.dataframe(df, p_show)

        # s-attributes
        for s_att in s_show:
            tmp = self.corpus.dump2satt(df, s_att)
            if s_att in tmp.columns:  # annotation
                df[s_att] = tmp[s_att]
            else:               # ID if no annotation
                df[s_att] = tmp[s_att + "_cwbid"]
            if cwb_ids:
                df[s_att + "_cwbid"] = tmp[s_att + "_cwbid"]

        return df


# def format_concordance(corpus, matches_df, p_show, s_show, order, cut_off, window, matches_filter, matches_highlight):

#     # TODO: simplify, retrieve more tokens left and right
#     concordance = Concordance(corpus, matches_df)
#     lines = concordance.lines(form='dict', p_show=p_show, s_show=s_show, order=order, cut_off=None)
#     out = list()
#     for line in lines.iterrows():

#         meta = {s: line[1][s] for s in s_show}
#         line = line[1]['dict']

#         roles = list()
#         discoursemes_in_window = {disc_name: False for disc_name in matches_filter.keys()}

#         for cpos, offset in zip(line['cpos'], line['offset']):
#             cpos_roles = list()
#             # node
#             if offset == 0:
#                 cpos_roles.append('node')
#             # out of window
#             elif abs(offset) > window:
#                 cpos_roles.append('out_of_window')

#             # highlighting
#             for disc_name, disc_matches in matches_highlight.items():
#                 if cpos in disc_matches:
#                     cpos_roles.append(disc_name)

#             # filtering
#             for disc_name, disc_matches in matches_filter.items():
#                 if cpos in disc_matches:
#                     cpos_roles.append(disc_name)
#                     if abs(offset) <= window:
#                         discoursemes_in_window[disc_name] = True

#             roles.append(cpos_roles)

#         # we filter here according to window size
#         if sum(discoursemes_in_window.values()) >= len(matches_filter):
#             line['lemma'] = [cqp_escape(item) for item in line['lemma']]
#             line['role'] = roles
#             line['meta'] = DataFrame.from_dict(meta, orient='index').to_html(bold_rows=False, header=False, render_links=True)
#             out.append(line)

#         if len(out) >= cut_off:
#             break

#     return out


def format_line(corpus, index, row, p_show, s_show, matches_filter, matches_highlight, window, htmlify_meta=True, cwb_ids=False):

    # pack all values into a dictionary
    row = dict(row)
    match, matchend = index
    row['match'] = match
    row['matchend'] = matchend

    # create cotext
    cotext = _node2cotext(row['match'], row['matchend'], row['context'], row['contextend'])

    # init output dictionary
    d = {
        'match': match,
        'cpos': cotext['cpos_list'],
        'offset': cotext['offset_list']
    }

    # role formatting and window-based cut-off
    roles = list()
    discoursemes_in_window = {disc_name: False for disc_name in matches_filter.keys()}

    for cpos, offset in zip(d['cpos'], d['offset']):

        cpos_roles = list()

        # node
        if offset == 0:
            cpos_roles.append('node')
        # out of window
        elif abs(offset) > window:
            cpos_roles.append('out_of_window')

        # highlighting
        for disc_name, disc_matches in matches_highlight.items():
            if cpos in disc_matches:
                cpos_roles.append(disc_name)

        # filtering
        for disc_name, disc_matches in matches_filter.items():
            if cpos in disc_matches:
                cpos_roles.append(disc_name)
                if abs(offset) <= window:
                    discoursemes_in_window[disc_name] = True

        roles.append(cpos_roles)

    # we filter here according to window size
    if sum(discoursemes_in_window.values()) >= len(matches_filter):

        # lexicalize cpos
        p_att_lists = zip(*list(map(lambda x: corpus.cpos2patts(x, p_show), d['cpos'])))
        for p, a in zip(p_show, p_att_lists):
            d[p] = list(a)

        # retrieve meta data at match
        meta = dict()
        for s in s_show:
            s_att = corpus.attributes.attribute(s, 's')
            try:
                meta[s] = decode(s_att[s_att.cpos2struc(match)][2])
                if cwb_ids:
                    meta[s + "_cwbid"] = s_att.cpos2struc(match)
            except KeyError:
                meta[s] = None

        # add meta data
        if htmlify_meta:
            d['meta'] = DataFrame.from_dict(meta, orient='index').to_html(bold_rows=False, header=False, render_links=True)
        else:
            for s, m in meta.items():
                d[s] = m

        # add roles
        d['role'] = roles

        # post-process lemmata
        if 'lemma' in d.keys():
            d['lemma'] = [cqp_escape(item) for item in d['lemma']]

        return d

    else:
        return None



#cython: cdivision=True

cpdef double choose(int n, int k):
    """
    Binomial coefficient Cython implementation
    :param int n: n elements
    :param int k: k subset elements
    :rtype: int
    """

    if k < 0:
        return 0
    if k == 0:
        return 1
    if n < k:
        return 0

    cdef double p = 1
    cdef int N = min(k, n - k) + 1
    cdef int i
    for i in range(1, N):
        p *= n
        p = int(p / i)
        n -= 1
    return p


from ccc import Corpus
from ccc.cwb import SubCorpora
from .conftest import DATA_PATH
import pytest


def test_subcorpora_from_dataframe(germaparl):

    corpus = Corpus(
        germaparl['corpus_name'],
        registry_dir=germaparl['registry_dir'],
        lib_dir=None,
        data_dir=DATA_PATH
    )

    factions = corpus.query_s_att('text_parliamentary_group').df.drop('text_parliamentary_group_cwbid', axis=1)
    factions.columns = ['subcorpus']

    subcorpora = SubCorpora(
        germaparl['corpus_name'],
        factions,
        registry_dir=germaparl['registry_dir'],
        lib_dir=None,
        data_dir=DATA_PATH
    )

    assert isinstance(subcorpora, SubCorpora)


@pytest.mark.now
def test_subcorpora_query(germaparl):

    corpus = Corpus(
        germaparl['corpus_name'],
        registry_dir=germaparl['registry_dir'],
        lib_dir=None,
        data_dir=DATA_PATH
    )

    factions = corpus.query_s_att('text_parliamentary_group').df.drop('text_parliamentary_group_cwbid', axis=1)
    factions.columns = ['subcorpus']

    subcorpora = SubCorpora(
        germaparl['corpus_name'],
        factions,
        registry_dir=germaparl['registry_dir'],
        lib_dir=None,
        data_dir=DATA_PATH
    )

    subcorpora.query('[lemma="und"]')

    print(subcorpora.nodes)


@pytest.mark.now
def test_subcorpora_collocates(germaparl):

    corpus = Corpus(
        germaparl['corpus_name'],
        registry_dir=germaparl['registry_dir'],
        lib_dir=None,
        data_dir=DATA_PATH
    )

    factions = corpus.query_s_att('text_parliamentary_group').df.drop('text_parliamentary_group_cwbid', axis=1)
    factions.columns = ['subcorpus']

    subcorpora = SubCorpora(
        germaparl['corpus_name'],
        factions,
        registry_dir=germaparl['registry_dir'],
        lib_dir=None,
        data_dir=DATA_PATH
    )

    subcorpora.query('[lemma="und"]')

    print(subcorpora.collocates())


@pytest.mark.now
def test_subcorpora_concordance(germaparl):

    corpus = Corpus(
        germaparl['corpus_name'],
        registry_dir=germaparl['registry_dir'],
        lib_dir=None,
        data_dir=DATA_PATH
    )

    factions = corpus.query_s_att('text_parliamentary_group').df.drop('text_parliamentary_group_cwbid', axis=1)
    factions.columns = ['subcorpus']

    subcorpora = SubCorpora(
        germaparl['corpus_name'],
        factions,
        registry_dir=germaparl['registry_dir'],
        lib_dir=None,
        data_dir=DATA_PATH
    )

    subcorpora.query('[lemma="und"]')

    print(subcorpora.concordance())
